import {file_url, system_file, nodes, is_mac, zip_file} from "../nodes.js";

import {NaiveUiResolver} from "unplugin-vue-components/dist/resolvers.js";

import Components from 'unplugin-vue-components/vite';

import viteCompression from "vite-plugin-compression";

import AutoImport from "unplugin-auto-import/vite";

import {createHtmlPlugin} from "vite-plugin-html";

import Markdown from 'unplugin-vue-markdown/vite';

import {viteMockServe} from "vite-plugin-mock";

import setView from "../route/setVue.js";

import viteVue from "@vitejs/plugin-vue";

import svgLoader from "vite-svg-loader";

import alone from "../alone.js";

import path from "path";

const plugin_transform = function (code, file, main, before, after = '') {
    if (!before) return null;
    const trim_rep = (path) => path.replace(/\\/g, '/').replace(/^\/+|\/+$/g, '');
    const code_arr = trim_rep(file).split('/');
    const main_file = trim_rep(main);
    const main_arr = main_file.split('/');
    if (code_arr.slice(-main_arr.length).join('/') != main_file) return null;
    const new_code = (typeof before == 'function') ? before(code) : (before + code + after);
    return {code: new_code, map: null};
};

const getConfig = async function (pathname, option = {}) {
    const self = {
        //配置
        option: Object.assign({
            //模式
            mode: "development",
            //打包还是服务
            command: "build",
            //用户根目录 绝对路径
            pathname: process.cwd(),
            //开发目录相对路径
            path: (pathname || "src"),
            //配置文件名
            conf: "config/conf.js",
            //js文件名
            env: "config/env.js",
            //env.js开发key
            dev_key: "_dev_",
            //App.vue文件名
            app: "App.vue",
            //main.js文件名
            main: "main.js",
            //当前根目录 绝对路径
            dirname: file_url(import.meta.url, 2),
            //临时信息
            cache: "file/cache.json",
            //复制的目录 绝对路径
            copy_dir: alone.rtrim(file_url(import.meta.url).replace(/\\/g, '/'), '/') + "/src",
        }, option || {}),
        //判断是否/开头
        is_run(path) {
            return path.substring(0, 1) == '/';
        },
        //\替换成/
        rep(file) {
            return file.replace(/\\/g, '/');
        },
        trim_rep(file) {
            return alone.trim(this.rep(file), '/');
        },
        rtrim_rep(file) {
            return alone.rtrim(this.rep(file), '/');
        },
        //加载文件
        async import_file(file) {
            return await import(system_file(file));
        },
        //打包设置
        build: {
            //文件格式
            format: {
                img: ['jpeg', 'jpg', 'png', 'gif', 'svg', 'svgz', 'bmp', 'tiff', 'webp', 'heif', 'heic', 'ico'],
                font: ['ttf', 'otf', 'woff', 'woff2', 'eot'],
                audio: ['mp3', 'wav', 'ogg', 'aac', 'flac'],
                video: ['mp4', 'webm', 'ogv', 'avi', 'mov']
            },
            //文件格式 array
            format_arr: [],
            //打包配置
            config: {
                //格式 包含有'svg', 'svgz',生成vendor.svg, svg或者file体积太大时可以选择开启
                svg: ['svg', 'svgz'],
                //路径包含有/vue-,生成vendor.vue, chunk体积太大时可以选择开启
                vue: (file, format) => {
                    if (file.includes('/vue-') && format.includes('js')) {
                        return true;
                    }
                },
                //路径包含有/alone/,生成vendor.alone
                alone: (file, format) => {
                    if (file.includes('/alone') && format.includes('js')) {
                        return true;
                    }
                },
                //路径包含有/naive-ui,生成vendor.naive, 推荐开启
                naive: (file, format) => {
                    if (file.includes('/naive-ui') && format.includes('js')) {
                        return true;
                    }
                },
                //格式 包含图片,字体,音频,视频打包成 vendor.file, 推荐开启
                file: (file, format) => {
                    if (self.build.format_arr.includes(format)) {
                        return true;
                    }
                },
                //路径包含有node_modules helper 打包成vendor.chunk, 推荐开启
                chunk: (file, format) => {
                    const chunk = ['node_modules', 'helpers', '-helper'];
                    if (chunk.some(val => file.includes(val))) {
                        return true;
                    }
                }
            },
        },
        //jsconfig.json配置
        js_config(object, path, pub, view, paths = {}) {
            paths["~/*"] = ["./*"];
            paths["&/*"] = ["./" + alone.trim(pub, '/') + "/*", "./" + path + "/" + alone.trim(pub, '/') + "/*"];
            paths["%/*"] = ["./" + path + "/" + view + "/*"];
            paths["@/*"] = ["./" + path + "/*"];
            object.compilerOptions = object.compilerOptions || {};
            object.compilerOptions.paths = paths;
            object.include = ["./" + path + "/**/*"];
            return object;
        },
        //1
        handle_before() {
            //是否更新
            this.update = false;
            //文件格式 array
            this.build.format_arr = [].concat(...Object.values(this.build.format));
            //是否开发模式
            this.is_mode = this.option.mode == 'development';
            //是否build
            this.is_build = this.option.command == 'build';
            //当前根目录 node
            this.vite = nodes(this.option.dirname);
            //用户根目录 node
            this.node = nodes(this.option.pathname);
            //复制目录 node
            this.copy = nodes(this.option.copy_dir);
            //开发目录
            this.app_path = this.trim_rep(this.option.path);
            //开发目录 node
            this.app = nodes(this.node.get_path(this.app_path));
            return this;
        },
        //2
        async handle_env() {
            this.option.env = this.trim_rep(this.option.env);
            if (!this.app.is_exist(this.option.env)) {
                this.app.cp_file(this.copy.get_path("config/env.js"), this.option.env, false);
            }
            this.env = {};
            this.env_data = (await this.import_file(this.app.get_path(this.option.env))).default || {};
            Object.keys(this.env_data).forEach((k) => {
                if (this.is_mode) {
                    const key = this.option.dev_key + k;
                    if (this.env_data.hasOwnProperty(key)) {
                        this.env[k] = this.env_data[key];
                    } else if (k.substring(0, this.option.dev_key.length) == this.option.dev_key) {
                        this.env[k.substring(this.option.dev_key.length)] = this.env_data[k];
                    } else if (!this.env.hasOwnProperty(k)) {
                        this.env[k] = this.env_data[k];
                    }
                } else if (k.substring(0, this.option.dev_key.length) != this.option.dev_key) {
                    this.env[k] = this.env_data[k];
                }
            });
            this.env.mock = (this.is_mode && this.env.mock) ? this.app.get_path('mock') : "";
            this.env.view = this.trim_rep(this.env.view || 'view');
            this.env.public = this.trim_rep(this.env.public || 'public');
            this.env.build = this.env.build || {};
            return this;
        },
        //3
        cache_file() {
            this.cache = this.node.get_json(this.option.cache);
            if (!alone.isset(this.cache) || this.cache.path != this.app_path || this.cache.view != this.env.view) {
                this.update = true;
                this.vite.save_json(this.option.cache, {path: this.app_path, view: this.env.view});
            }
            return this;
        },
        //4
        jsconfig() {
            this.is_jsconfig = this.node.is_exist('jsconfig.json');
            if (this.update || !this.is_jsconfig) {
                this.jsconfig_object = !this.is_jsconfig ? this.vite.get_json('jsconfig.json') : this.node.get_json('jsconfig.json');
                this.jsconfig_object = this.js_config(this.jsconfig_object, this.app_path, this.env.public, this.env.view);
                this.node.save_json('jsconfig.json', this.jsconfig_object);
            }
            return this;
        },
        //5
        tsconfig() {
            this.is_tsconfig = this.node.is_exist('tsconfig.json');
            if (!this.is_tsconfig) {
                this.node.save_json('tsconfig.json', this.vite.get_json('tsconfig.json'));
            }
            return this;
        },
        //6
        copy_app() {
            if (!this.app.is_exist(this.option.app)) {
                this.update = true;
                this.app.cp_file(this.copy.get_path("App.vue"), this.option.app, false);
                this.app.cp_file(this.copy.get_path("main.js"), this.option.main, false);
                this.app.cp_file(this.copy.get_path("config/conf.js"), this.option.conf, false);
                this.app.cp_dir_file(this.copy.get_path("view"), this.env.view, [], false);
                this.app.cp_dir_file(this.copy.get_path("mock"), 'mock', [], false);
                this.app.cp_dir_file(this.copy.get_path("style"), 'style', [], false);
            }
            if (!this.node.is_exist(this.env.public + "/layui/layui.js")) {
                this.node.cp_dir_file(this.vite.get_path("style/layui"), this.env.public + "/layui", [], false);
                this.node.cp_file(this.vite.get_path("style/lay.js"), this.env.public + "/layui/lay.js", false);
            }
            if (this.app.is_exist(this.env.public)) {
                this.node.cp_dir_file(this.app.get_path(this.env.public), this.env.public, false);
            }
            return this;
        },
        //7
        handle_dist() {
            if (alone.isset(this.env.dist)) {
                this.out_dir = this.is_run(this.env.dist) ? this.env.dist : this.node.get_path(this.trim_rep(this.env.dist));
            } else {
                this.dist = "dist/" + (this.app_path.split('/').length > 1 ? (this.app_path.split('/').slice(1).join('/')) : this.app_path);
                this.out_dir = this.node.get_path(this.dist);
            }
            return this;
        },
        //8
        handle_favicon() {
            if (this.env.favicon && !this.is_run(this.env.favicon)) {
                this.env.favicon = alone.trim(alone.trim(this.env.favicon, '.'), '/');
                this.env.favicon = this.app.is_exist(this.env.favicon) ? (this.app_path + "/" + this.env.favicon) : "";
            }
            return this;
        },
        //9
        async handle_boj() {
            this.env_obj = {...this.env};
            this.conf_obj = (this.app.is_exist(this.option.conf) ? ((await this.import_file(this.app.get_path(this.option.conf))).default) : {}) || {};
            this.del_env = ['host', 'port', 'build', 'zip', 'public', 'mock', 'favicon', 'view', 'dist', 'layui'];
            this.del_env.forEach((k) => ((this.env_obj.hasOwnProperty(k)) && delete this.env_obj[k]));
            return this;
        },
        //10
        handle_chunk() {
            this.chunk = {svg: false, vue: false, alone: true, naive: true, file: true, chunk: true};
            Object.keys(this.build.config).forEach((k) => {
                if (this.env.build[k] === true) {
                    this.chunk[k] = this.build.config[k];
                } else if (alone.isset(this.env.build[k])) {
                    this.chunk[k] = this.env.build[k];
                } else if (!this.env.build.hasOwnProperty(k) && this.chunk[k] === true) {
                    this.chunk[k] = this.build.config[k];
                } else {
                    delete this.chunk[k];
                }
            });
            if (alone.isset(this.env.build.callback) && alone.is_object(this.env.build.callback)) {
                Object.keys(this.env.build.callback).forEach((key) => {
                    if (alone.isset(this.env.build.callback[key])) {
                        this.chunk[key] = this.env.build.callback[key];
                    }
                });
            }
            return this;
        },
        //11
        handle_html_tag() {
            this.lay_def = {status: true, link: "/layui/css/layui.css", script: "/layui/layui.js"};
            this.html_tag = [
                {tag: 'meta', attrs: {charset: 'UTF-8'}, injectTo: 'head-prepend'},
                {
                    tag: 'meta', attrs: {name: 'viewport', content: 'width=device-width, initial-scale=1.0'},
                    injectTo: 'head-prepend',
                },
                {
                    tag: 'div',
                    attrs: {id: "app"},
                    injectTo: 'body',
                    children: `<style>body{margin:0}.sk-flow{width:50px;height:50px;display:flex;justify-content:space-between}.sk-flow-dot{width:25%;height:25%;background-color:#79bbff;border-radius:50%;animation:sk-flow 1.4s cubic-bezier(.455,.03,.515,.955)0s infinite both}.sk-flow-dot:first-child{animation-delay:-.3s}.sk-flow-dot:nth-child(2){animation-delay:-.15s}.sk-main{width:100vw;height:100vh;display:flex;align-items:center;justify-content:center}@keyframes sk-flow{0%,100%,80%{transform:scale(.3)}40%{transform:scale(1)}}</style><div class="sk-main"><div class="sk-flow"><div class="sk-flow-dot"></div><div class="sk-flow-dot"></div><div class="sk-flow-dot"></div></div></div>`,
                }
            ];
            this.env.layui = this.env.layui || this.lay_def;
            if (this.env.layui.status !== false) {
                this.app_ver = alone.rand_str();
                this.html_tag = this.html_tag.concat([
                    {
                        tag: 'link',
                        attrs: {
                            rel: 'stylesheet',
                            href: (this.env.layui.link || this.lay_def.link) + "?v=" + this.app_ver
                        },
                        injectTo: 'head',
                    },
                    {
                        tag: 'script',
                        attrs: {
                            type: 'module',
                            src: (this.env.layui.script || this.lay_def.script) + "?v=" + this.app_ver
                        },
                        injectTo: 'head-prepend',
                    },
                    {
                        tag: 'script',
                        attrs: {type: 'module'},
                        injectTo: 'head',
                        children: `window.$ = layui.$;`,
                    },
                ]);
            }
            if (this.env.favicon) {
                this.html_tag = this.html_tag.concat([{
                    tag: 'link',
                    attrs: {icon: 'favicon', rel: "icon", href: this.env.favicon},
                    injectTo: 'head',
                }]);
            }
        },
        //启动
        async start() {
            this.handle_before();
            await this.handle_env();
            this.cache_file();
            this.jsconfig();
            this.tsconfig();
            this.copy_app();
            this.handle_dist();
            this.handle_favicon();
            await this.handle_boj();
            this.handle_chunk();
            this.handle_html_tag();
            return this;
        },
        publicDir() {
            return this.node.get_path(this.env.public);
        },
        define() {
            return {'process.env': this.env_obj, 'process.config': this.conf_obj};
        },
        alias() {
            return {
                '~': this.node.get_path('.'),
                '&': this.node.get_path(this.env.public),
                '@': this.app.get_path('.'),
                '%': this.app.get_path(this.env.view)
            };
        },
        preview() {
            return {host: this.env.host || '127.0.0.1', port: this.env.port || '8088'};
        },
        server() {
            return Object.assign({
                proxy: {
                    '/api': {
                        target: this.env.api,
                        changeOrigin: true,
                        rewrite: (path) => path.replace(/^\/api/, '')
                    }
                }
            }, this.preview());
        },
        vue_route({dir, file} = {}) {
            setView(this.app.get_path(this.env.view), {dir: dir, file: file, run: true});
        },
        plugin_transform_main(code, file, before, after = '') {
            return plugin_transform(code, file, this.app_path + "/" + this.option.main, before, after);
        },
        plugin_build() {
            return {
                outDir: this.out_dir,
                target: 'esnext',//'esnext', es2015 es2019
                minify: 'terser',//'terser',esbuild
                cssCodeSplit: true, // 如果设置为false，整个项目中的所有 CSS 将被提取到一个 CSS 文件中
                reportCompressedSize: false,//  关闭生成map文件 可以达到缩小打包体积
                chunkSizeWarningLimit: 1500,
                sourcemap: false, // 这个生产环境一定要关闭，不然打包的产物会很大
                terserOptions: {
                    format: {
                        beautify: false, // 不格式化输出
                        comments: false, // 不保留注释
                    },
                    mangle: {
                        toplevel: true, // 混淆顶级变量和函数
                    },
                    compress: {
                        properties: true,
                        dead_code: true,    // 删除未使用的代码
                        drop_console: true,//打包时删除console
                        drop_debugger: true,//打包时删除 debugger
                    },
                    sourceMap: true, // 生成源映射文件
                    ecma: 2015,      // 使用 ECMAScript 2015 的特性
                    enclose: false,  // 不包裹代码在闭包中
                    keep_classnames: false, // 不保留类名
                    keep_fnames: false, // 不保留函数名
                    ie8: false,       // 不兼容 IE8
                    module: false,    // 不支持 ES6 模块语法
                    nameCache: null,  // 不使用名称缓存
                    safari10: false,  // 不兼容 Safari 10
                    toplevel: false,  // 不混淆顶级作用域的变量
                },
                rollupOptions: {
                    treeshake: true, // 开启 Tree Shaking，消除未使用的代码，减小最终的包大小
                    output: {
                        hashCharacters: 'hex',
                        chunkFileNames: (obj) => {
                            const name = (obj.name).toString().toLowerCase()
                            return 'assets/js/' + name + '.[hash].js';
                        },
                        entryFileNames: (obj) => {
                            const name = (obj.name).toString().toLowerCase()
                            return 'assets/js/' + name + '.[hash].js';
                        },
                        assetFileNames: (obj) => {
                            const type = obj.type.toString().toLowerCase();
                            if (type === 'asset' && (obj.name || obj.names)) {
                                const file = (obj.name || (obj.names.slice(0, 1).join(''))).toString();
                                const files = file.replace(/\\/g, '/').toLowerCase();
                                const format = path.extname(files).slice(1);
                                const array = Object.keys(this.build.format);
                                for (const key of array) {
                                    const val = this.build.format[key];
                                    if (val && ((typeof val == 'function') ? val(files, format) : (Array.isArray(val) ? val.includes(format) : files.includes(val)))) {
                                        return 'assets/' + key + '/[name].[hash].[ext]';
                                    }
                                }
                            }
                            return 'assets/[ext]/[name].[hash].[ext]';
                        },
                        manualChunks: (file) => {
                            const files = file.replace(/\\/g, '/').toLowerCase();
                            const format = path.extname(files).slice(1);
                            const array = Object.keys(this.chunk);
                            for (const key of array) {
                                const val = this.chunk[key];
                                if (val && ((typeof val == 'function') ? val(files, format) : (Array.isArray(val) ? val.includes(format) : files.includes(val)))) {
                                    return "vendor." + key;
                                }
                            }
                        }
                    }
                }
            };
        }
    };
    return (await self.start());
};

export default async function (pathname, option = {}, {
    vue = {},
    html = {},
    md = {},
    mock = {},
    svg = {},
    compress = {},
    auto = {},
    component = {}
} = {}) {
    const then = await getConfig(pathname, option);
    const plugin = {
        vue(data) {
            return viteVue(alone.array_merge({include: [/\.vue$/, /\.md$/]}, data));
        },
        compress(data) {
            return viteCompression(alone.array_merge({
                verbose: true,    // 默认即可
                disable: true,  //开启压缩(不禁用)，默认即可
                deleteOriginFile: false, //删除源文件
                threshold: 10240, //压缩前最小文件大小
                algorithm: 'gzip',  //压缩算法
                ext: '.gz', //文件类型
            }, data));
        },
        svg(data) {
            return svgLoader(alone.array_merge({defaultImport: 'url'}, data));
        },
        html(data) {
            const tags = alone.get_arr(data, 'inject.tags');
            const config = alone.array_merge(data, {minify: true, inject: {tags: []}});
            config.inject.tags = (alone.isset(tags) && alone.is_array(tags)) ? then.html_tag.concat(tags) : then.html_tag;
            config.entry = then.app_path + "/" + then.option.main;
            return createHtmlPlugin(config);

        },
        md(data) {
            const config = {headEnabled: true, markdownItOptions: {html: true, linkify: true, typographer: true}};
            const include = [`${then.node.get_path('.')}/*.md`, `${then.node.get_path('.')}/**/*.md`];
            const includes = alone.get_arr(data, 'include');
            config.include = (alone.isset(includes) && alone.is_array(includes)) ? include.concat(includes) : include;
            return Markdown(config);
        },
        mock(data) {
            return viteMockServe(alone.array_merge({
                mockPath: then.env.mock,//存放 Mock 文件的路径。默认情况下，插件会在此路径下查找 Mock 文件
                watchFiles: true,//指示是否监视 Mock 文件的变化。如果设置为 true，当 Mock 文件发生变化时，Mock 服务器会自动更新。
                enable: true,//指示是否启用 Mock 功能。设置为 false 可以临时禁用 Mock。
                logger: false,//指示是否启用插件的日志记录。设置为 true 会在终端输出 Mock 请求的日志信息。
                cors: true //指示是否启用 CORS（跨域资源共享）。设置为 true 会在响应中添加 CORS 头部，以允许跨域请求。
            }, data || {}));
        },
        auto(data) {
            return AutoImport(alone.array_merge({
                imports: ['vue', {
                    'naive-ui': [
                        'useDialog',
                        'useMessage',
                        'useNotification',
                        'useLoadingBar'
                    ]
                }]
            }, data || {}));
        },
        component(data) {
            return Components(alone.array_merge({
                extensions: ['vue', 'md'],
                include: [/\.vue$/, /\.vue\?vue/, /\.md$/],
                resolvers: [NaiveUiResolver()],
            }, data || {}));
        },
        alone() {
            return {
                enforce: 'pre',
                name: 'alone-frame',
                config(config) {

                    config.base = config.base || "./";

                    config.publicDir = then.publicDir();

                    config.resolve = config.resolve || {};

                    config.define = alone.array_merge(config.define, then.define());

                    config.resolve.alias = alone.array_merge(config.resolve.alias, then.alias());

                    config.build = alone.array_merge(config.build, then.plugin_build());

                    config.preview = alone.array_merge(config.preview, then.preview());

                    config.server = alone.array_merge(config.server, then.server());
                },
                buildStart() {
                    then.vue_route();
                    console.log('path', `${is_mac() ? "file://" : ""}${then.app.get_path('.')}`);
                },
                closeBundle() {
                    if (then.is_build) {
                        if (then.env.zip === true) {
                            zip_file(then.out_dir, alone.rtrim(then.out_dir, '/') + ".zip", function (size, unit, file) {
                                console.log('zip file', file);
                                console.log('zip size', `${size}${unit}`);
                            });
                        }
                    }
                    console.log('complete', `${is_mac() ? "file://" : ""}` + then.out_dir);
                },
                transform(code, file) {
                    const beforeCode = `
            import MainLoadBefore from "alone-frame/frame/main";
            MainLoadBefore.status.load = false;
            await MainLoadBefore.load();
            `;
                    const afterCode = `
            import MainLoadAfter from "alone-frame/frame/main";
            MainLoadAfter.status.mount = false;
            MainLoadAfter.mount();
            `;
                    return then.plugin_transform_main(code, file, beforeCode, afterCode);
                }
            };
        }
    };
    return [
        (vue !== false && plugin.vue(vue)),
        plugin.alone(),
        (compress !== false && plugin.compress(compress)),
        (svg !== false && plugin.svg(svg)),
        (html !== false && plugin.html(html)),
        (md !== false && plugin.md(md)),
        ((then.env.mock && mock !== false) && plugin.mock(mock)),
        (auto !== false && plugin.auto(auto)),
        (component !== false && plugin.component(component))
    ];
}